@@ -33,6 +33,7 @@ import org.apache.hadoop.fs.Path;
 import org.apache.sysds.api.DMLScript;
 import org.apache.sysds.api.jmlc.JMLCUtils;
 import org.apache.sysds.common.Types.DataType;
+import org.apache.sysds.common.Types.ExecType;
 import org.apache.sysds.common.Types.FileFormat;
 import org.apache.sysds.common.Types.OpOp1;
 import org.apache.sysds.common.Types.OpOpDG;
@@ -57,7 +58,6 @@ import org.apache.sysds.hops.codegen.SpoofCompiler;
 import org.apache.sysds.hops.rewrite.HopRewriteUtils;
 import org.apache.sysds.hops.rewrite.ProgramRewriter;
 import org.apache.sysds.lops.Lop;
-import org.apache.sysds.common.Types.ExecType;
 import org.apache.sysds.lops.compile.Dag;
 import org.apache.sysds.parser.DMLProgram;
 import org.apache.sysds.parser.DataExpression;
@@ -93,6 +93,7 @@ import org.apache.sysds.runtime.instructions.cp.IntObject;
 import org.apache.sysds.runtime.instructions.cp.ListObject;
 import org.apache.sysds.runtime.instructions.cp.ScalarObject;
 import org.apache.sysds.runtime.io.IOUtilFunctions;
+import org.apache.sysds.runtime.lineage.LineageItem;
 import org.apache.sysds.runtime.matrix.data.MatrixBlock;
 import org.apache.sysds.runtime.meta.DataCharacteristics;
 import org.apache.sysds.runtime.meta.MatrixCharacteristics;
@@ -114,8 +115,9 @@ import org.apache.sysds.utils.Explain.ExplainType;
  * 
  * 
  */
-public class Recompiler 
-{
+public class Recompiler {
+	// private static final Log LOG =  LogFactory.getLog(Recompiler.class.getName());
+
 	//Max threshold for in-memory reblock of text input [in bytes]
 	//reason: single-threaded text read at 20MB/s, 1GB input -> 50s (should exploit parallelism)
 	//note that we scale this threshold up by the degree of available parallelism
@@ -1323,6 +1325,10 @@ public class Recompiler
 					d.setDim1(mo.getNumRows());
 					d.setDim2(mo.getNumColumns());
 					d.setNnz(mo.getNnz());
+					if(mo.isCompressed()){
+						d.setCompressedSize(mo.getCompressedSize());
+						d.setCompressedOutput(true);	
+					}
 				}
 				else if( dat instanceof FrameObject ) {
 					FrameObject fo = (FrameObject) dat;
@@ -1341,6 +1347,10 @@ public class Recompiler
 					d.setDim2(to.getNumColumns());
 					d.setNnz(to.getNnz());
 				}
+				if( dat instanceof CacheableData<?> ) {
+					CacheableData<?> cd = (CacheableData<?>) dat;
+					d.setOnlyRDD(!cd.isCached(true) &&cd.getRDDHandle()!=null);
+				}
 			}
 		}
 		//special case for persistent reads with unknown size (read-after-write)
@@ -1569,7 +1579,7 @@ public class Recompiler
 				throw new DMLRuntimeException(ex);
 			}
 		}
-		
+
 		//check valid dimensions and memory requirements
 		double sp = OptimizerUtils.getSparsity(rows, cols, nnz);
 		double mem = MatrixBlock.estimateSizeInMemory(rows, cols, sp);
@@ -1584,7 +1594,8 @@ public class Recompiler
 		long estFilesize = (long)(3.5 * mem); //conservative estimate
 		long cpThreshold = CP_REBLOCK_THRESHOLD_SIZE * 
 			OptimizerUtils.getParallelTextReadParallelism();
-		return (estFilesize < cpThreshold);
+		return (iimd.getFileFormat() == FileFormat.BINARY
+			|| estFilesize < cpThreshold); //for text conservative
 	}
 	
 	public static boolean checkCPCheckpoint(DataCharacteristics dc) {
@@ -1592,9 +1603,13 @@ public class Recompiler
 			&& OptimizerUtils.isValidCPDimensions(dc.getRows(), dc.getCols())
 			&& !OptimizerUtils.exceedsCachingThreshold(dc.getCols(), OptimizerUtils.estimateSize(dc));
 	}
+
+	public static void executeInMemoryReblock(ExecutionContext ec, String varin, String varout) {
+		executeInMemoryReblock(ec, varin, varout, null);
+	}
 	
 	@SuppressWarnings("unchecked")
-	public static void executeInMemoryReblock(ExecutionContext ec, String varin, String varout) {
+	public static void executeInMemoryReblock(ExecutionContext ec, String varin, String varout, LineageItem litem) {
 		CacheableData<CacheBlock> in = (CacheableData<CacheBlock>) ec.getCacheableData(varin);
 		CacheableData<CacheBlock> out = (CacheableData<CacheBlock>) ec.getCacheableData(varout);
 
@@ -1609,6 +1624,7 @@ public class Recompiler
 			
 			//set output (incl update matrix characteristics)
 			out.acquireModify(mb);
+			out.setCacheLineage(litem);
 			out.release();
 			in.release();
 		}
