@@ -19,36 +19,55 @@
 
 package org.apache.sysds.runtime.compress.colgroup.insertionsort;
 
+import org.apache.sysds.runtime.compress.DMLCompressionException;
 import org.apache.sysds.runtime.compress.colgroup.mapping.AMapToData;
 import org.apache.sysds.runtime.compress.colgroup.mapping.MapToFactory;
 import org.apache.sysds.runtime.compress.utils.IntArrayList;
 
 public class MaterializeSort extends AInsertionSorter {
 
+	/** The block size to materialize at a time */
+	public static int CACHE_BLOCK = 16000;
+
 	/** a dense mapToData, that have a value for each row in the input. */
-	final private AMapToData md;
-	final private int CACHE_BLOCK = 60000;
-	final private int[] skip;
+	private final AMapToData md;
+	private final int[] skip;
+
 	private int off = 0;
 
+	protected MaterializeSort(int endLength, int numRows, IntArrayList[] offsets) {
+		super(endLength, numRows, offsets);
+		// + 1 to ensure that the _numLabels is possible to represent in the map.
+		md = MapToFactory.create(Math.min(_numRows, CACHE_BLOCK), _numLabels + 1);
+		skip = new int[offsets.length];
+		for(int block = 0; block < _numRows; block += CACHE_BLOCK)
+			insert(block, Math.min(block + CACHE_BLOCK, _numRows));
+	}
+
 	protected MaterializeSort(int endLength, int numRows, IntArrayList[] offsets, int negativeIndex) {
 		super(endLength, numRows, offsets, negativeIndex);
-
-		md = MapToFactory.create(Math.min(_numRows, CACHE_BLOCK), _numLabels);
+		// + 1 to ensure that the _numLabels is possible to represent in the map.
+		md = MapToFactory.create(Math.min(_numRows, CACHE_BLOCK), _numLabels + 1);
 		skip = new int[offsets.length];
-
-		for(int block = 0; block < _numRows; block += CACHE_BLOCK) {
-			md.fill(_numLabels);
-			if(_negativeIndex == -1)
-				insert(block, Math.min(block + CACHE_BLOCK, _numRows));
-			else
-				insertWithNegative(block, Math.min(block + CACHE_BLOCK, _numRows));
-		}
+		for(int block = 0; block < _numRows; block += CACHE_BLOCK)
+			insertWithNegative(block, Math.min(block + CACHE_BLOCK, _numRows));
 	}
 
 	private void insert(int rl, int ru) {
-		materializeInsert(rl, ru);
-		filterInsert(rl, ru);
+		try {
+			md.fill(_numLabels);
+			materializeInsert(rl, ru);
+			filterInsert(rl, ru);
+		}
+		catch(Exception e) {
+			int sum = 0;
+			for(IntArrayList o : _offsets)
+				sum += o.size();
+			throw new DMLCompressionException(
+				"Failed normal materialize sorting with list of " + _offsets.length + " with sum (aka output size): " + sum
+					+ " requested Size: " + _indexes.length + " range: " + rl + " " + ru,
+				e);
+		}
 	}
 
 	private void materializeInsert(int rl, int ru) {
@@ -63,14 +82,17 @@ public class MaterializeSort extends AInsertionSorter {
 	}
 
 	private void filterInsert(int rl, int ru) {
-		for(int i = rl; i < ru; i++) {
-			final int idx = md.getIndex(i - rl);
+		final int len = ru - rl;
+		for(int i = 0; i < len; i++) {
+			final int idx = md.getIndex(i);
 			if(idx != _numLabels)
-				set(off++, i, idx);
+				set(off++, i + rl, idx);
 		}
 	}
 
 	private void insertWithNegative(int rl, int ru) {
+		md.fill(_numLabels);
+
 		for(int i = 0; i < _offsets.length; i++) {
 			IntArrayList of = _offsets[i];
 			int k = skip[i];
