@@ -38,6 +38,7 @@ import org.apache.sysds.parser.StatementBlock;
 import org.apache.sysds.parser.VariableSet;
 import org.apache.sysds.runtime.DMLRuntimeException;
 import org.apache.sysds.runtime.controlprogram.caching.CacheableData;
+import org.apache.sysds.runtime.controlprogram.caching.FrameObject;
 import org.apache.sysds.runtime.controlprogram.caching.MatrixObject;
 import org.apache.sysds.runtime.controlprogram.context.ExecutionContext;
 import org.apache.sysds.runtime.controlprogram.context.SparkExecutionContext;
@@ -51,6 +52,7 @@ import org.apache.sysds.runtime.controlprogram.parfor.RemoteDPParForSpark;
 import org.apache.sysds.runtime.controlprogram.parfor.RemoteParForJobReturn;
 import org.apache.sysds.runtime.controlprogram.parfor.RemoteParForSpark;
 import org.apache.sysds.runtime.controlprogram.parfor.ResultMerge;
+import org.apache.sysds.runtime.controlprogram.parfor.ResultMergeFrameLocalMemory;
 import org.apache.sysds.runtime.controlprogram.parfor.ResultMergeLocalAutomatic;
 import org.apache.sysds.runtime.controlprogram.parfor.ResultMergeLocalFile;
 import org.apache.sysds.runtime.controlprogram.parfor.ResultMergeLocalMemory;
@@ -88,7 +90,7 @@ import org.apache.sysds.runtime.meta.DataCharacteristics;
 import org.apache.sysds.runtime.util.CollectionUtils;
 import org.apache.sysds.runtime.util.ProgramConverter;
 import org.apache.sysds.runtime.util.UtilFunctions;
-import org.apache.sysds.utils.Statistics;
+import org.apache.sysds.utils.stats.ParForStatistics;
 
 import java.io.Serializable;
 import java.util.ArrayList;
@@ -288,9 +290,9 @@ public class ParForProgramBlock extends ForProgramBlock
 	public static final boolean USE_RANGE_TASKS_IF_USEFUL   = true; // use range tasks whenever size>3, false, otherwise wrong split order in remote 
 	public static final boolean USE_STREAMING_TASK_CREATION = true; // start working while still creating tasks, prevents blocking due to too small task queue
 	public static final boolean ALLOW_NESTED_PARALLELISM    = true; // if not, transparently change parfor to for on program conversions (local,remote)
+	public static final boolean CONVERT_NESTED_REMOTE_PARFOR = true; //convert parfor to for in remote parfor
 	public static final boolean USE_PARALLEL_RESULT_MERGE   = false; // if result merge is run in parallel or serial 
 	public static final boolean USE_PARALLEL_RESULT_MERGE_REMOTE = true; // if remote result merge should be run in parallel for multiple result vars
-	public static final boolean ALLOW_DATA_COLOCATION       = true;
 	public static final boolean CREATE_UNSCOPED_RESULTVARS  = true;
 	public static       boolean ALLOW_REUSE_PARTITION_VARS  = true; //reuse partition input matrices, applied only if read-only in surrounding loops
 	public static final int     WRITE_REPLICATION_FACTOR    = 1;
@@ -318,19 +320,20 @@ public class ParForProgramBlock extends ForProgramBlock
 	protected final boolean _monitor;
 	protected final Level _optLogLevel;
 	protected int _numThreads = -1;
+	protected boolean _fixedDOP = false; //guard for numThreads
 	protected long _taskSize = -1;
-	protected PTaskPartitioner _taskPartitioner = null;
-	protected PDataPartitioner _dataPartitioner = null;
-	protected PResultMerge _resultMerge = null;
-	protected PExecMode _execMode = null;
-	protected POptMode _optMode = null;
+	protected PTaskPartitioner _taskPartitioner;
+	protected PDataPartitioner _dataPartitioner;
+	protected PResultMerge _resultMerge;
+	protected PExecMode _execMode;
+	protected POptMode _optMode;
 	
 	//specifics used for optimization
 	protected long _numIterations = -1;
 	
 	//specifics used for data partitioning
-	protected LocalVariableMap _variablesDPOriginal = null;
-	protected LocalVariableMap _variablesDPReuse = null;
+	protected LocalVariableMap _variablesDPOriginal;
+	protected LocalVariableMap _variablesDPReuse;
 	protected String _colocatedDPMatrix = null;
 	protected boolean _tSparseCol = false;
 	protected int _replicationDP = WRITE_REPLICATION_FACTOR;
@@ -460,6 +463,10 @@ public class ParForProgramBlock extends ForProgramBlock
 		return _optMode;
 	}
 	
+	public void setOptimizationMode(POptMode mode) {
+		_optMode = mode;
+	}
+	
 	public int getDegreeOfParallelism() {
 		return _numThreads;
 	}
@@ -469,6 +476,14 @@ public class ParForProgramBlock extends ForProgramBlock
 		_params.put(ParForStatementBlock.PAR, String.valueOf(_numThreads)); //kept up-to-date for copies
 		setLocalParWorkerIDs();
 	}
+	
+	public boolean isDegreeOfParallelismFixed() {
+		return _fixedDOP;
+	}
+	
+	public void setDegreeOfParallelismFixed(boolean flag) {
+		_fixedDOP = flag;
+	}
 
 	public void setCPCaching(boolean flag) {
 		_enableCPCaching = flag;
@@ -632,15 +647,15 @@ public class ParForProgramBlock extends ForProgramBlock
 			switch( _execMode )
 			{
 				case LOCAL: //create parworkers as local threads
-					executeLocalParFor(ec, iterVar, from, to, incr);
+					executeLocalParFor(ec, from, to, incr);
 					break;
 				
 				case REMOTE_SPARK: // create parworkers as Spark tasks (one job per parfor)
-					executeRemoteSparkParFor(ec, iterVar, from, to, incr);
+					executeRemoteSparkParFor(ec, from, to, incr);
 					break;
 				
 				case REMOTE_SPARK_DP: // create parworkers as Spark tasks (one job per parfor)
-					executeRemoteSparkParForDP(ec, iterVar, from, to, incr);
+					executeRemoteSparkParForDP(ec, from, to, incr);
 					break;
 				
 				default:
@@ -673,7 +688,7 @@ public class ParForProgramBlock extends ForProgramBlock
 		}
 		
 		//execute exit instructions
-		executeExitInstructions(_exitInstruction, "parfor", ec);
+		executeExitInstructions("parfor", ec);
 		
 		///////
 		//end PARALLEL EXECUTION of (PAR)FOR body
@@ -700,13 +715,12 @@ public class ParForProgramBlock extends ForProgramBlock
 	 * below for details of the realization.
 	 * 
 	 * @param ec execution context
-	 * @param itervar ?
 	 * @param from ?
 	 * @param to ?
 	 * @param incr ?
 	 * @throws InterruptedException if InterruptedException occurs
 	 */
-	private void executeLocalParFor( ExecutionContext ec, IntObject itervar, IntObject from, IntObject to, IntObject incr ) 
+	private void executeLocalParFor( ExecutionContext ec, IntObject from, IntObject to, IntObject incr )
 		throws InterruptedException
 	{
 		LOG.trace("Local Par For (multi-threaded) with degree of parallelism : " + _numThreads);
@@ -747,7 +761,7 @@ public class ParForProgramBlock extends ForProgramBlock
 			//maintain statistics
 			long tinit = (long) time.stop();
 			if( DMLScript.STATISTICS )
-				Statistics.incrementParForInitTime(tinit);
+				ParForStatistics.incrementInitTime(tinit);
 			if( _monitor ) 
 				StatisticMonitor.putPFStat(_ID, Stat.PARFOR_INIT_PARWRK_T, tinit);
 			
@@ -805,14 +819,18 @@ public class ParForProgramBlock extends ForProgramBlock
 				numExecutedIterations, numExecutedTasks, localVariables );
 			
 			// Step 5) cleanup local parworkers (e.g., remove created functions)
-			for( int i=0; i<_numThreads; i++ )
-			{
+			for( int i=0; i<_numThreads; i++ ) {
 				Collection<String> fnNames = workers[i].getFunctionNames();
-				if( fnNames!=null && !fnNames.isEmpty() )
-					for( String fn : fnNames ) {
-						String[] parts = DMLProgram.splitFunctionKey(fn);
-						_prog.removeFunctionProgramBlock(parts[0], parts[1]);
-					}
+				if( fnNames!=null ) 
+					fnNames.stream().map(fn -> DMLProgram.splitFunctionKey(fn))
+						.forEach(p -> _prog.removeFunctionProgramBlock(p[0], p[1]));
+				// also cleanup worker-specific functions created via eval on-demand loading
+				workers[i].getExecutionContext().getTmpParforFunctions().stream()
+					.map(fn -> DMLProgram.splitFunctionKey(fn))
+					.forEach(p -> {
+						_prog.getDMLProg().removeFunctionStatementBlock(p[0], p[1]);
+						_prog.removeFunctionProgramBlock(p[0], p[1]);
+					});
 			}
 
 			// Frees up the GPUContexts used in the threaded Parfor and sets
@@ -821,8 +839,7 @@ public class ParForProgramBlock extends ForProgramBlock
 				ec.getGPUContext(0).initializeThread();
 			}
 		}
-		finally 
-		{
+		finally {
 			//remove thread-local memory budget (reset to original budget)
 			//(in finally to prevent error side effects for multiple scripts in one jvm)
 			resetMemoryBudget();
@@ -835,7 +852,7 @@ public class ParForProgramBlock extends ForProgramBlock
 		}
 	}
 
-	private void executeRemoteSparkParFor(ExecutionContext ec, IntObject itervar, IntObject from, IntObject to, IntObject incr) 
+	private void executeRemoteSparkParFor(ExecutionContext ec, IntObject from, IntObject to, IntObject incr)
 	{
 		Timing time = ( _monitor ? new Timing(true) : null );
 		
@@ -901,7 +918,7 @@ public class ParForProgramBlock extends ForProgramBlock
 		}
 	}
 	
-	private void executeRemoteSparkParForDP( ExecutionContext ec, IntObject itervar, IntObject from, IntObject to, IntObject incr ) {
+	private void executeRemoteSparkParForDP( ExecutionContext ec, IntObject from, IntObject to, IntObject incr ) {
 		Timing time = ( _monitor ? new Timing(true) : null );
 		
 		// Step 0) check and compile to CP (if forced remote parfor)
@@ -976,7 +993,7 @@ public class ParForProgramBlock extends ForProgramBlock
 				Data dat = ec.getVariable(var);
 				//skip non-existing input matrices (which are due to unknown sizes marked for
 				//partitioning but typically related branches are never executed)
-				if( dat != null && dat instanceof MatrixObject )
+				if( dat instanceof MatrixObject )
 				{
 					MatrixObject moVar = (MatrixObject) dat; //unpartitioned input
 					
@@ -1056,12 +1073,12 @@ public class ParForProgramBlock extends ForProgramBlock
 	 * @param out output matrix
 	 * @param in array of input matrix objects
 	 */
-	private static void cleanWorkerResultVariables(ExecutionContext ec, MatrixObject out, MatrixObject[] in, boolean parallel) {
+	private static void cleanWorkerResultVariables(ExecutionContext ec, CacheableData<?> out, CacheableData<?>[] in, boolean parallel) {
 		//check for empty inputs (no iterations executed)
-		Stream<MatrixObject> results = Arrays.stream(in).filter(m -> m!=null && m!=out);
+		Stream<CacheableData<?>> results = Arrays.stream(in).filter(m -> m!=null && m!=out);
 		//perform cleanup (parallel to mitigate file deletion bottlenecks)
 		(parallel ? results.parallel() : results)
-			.forEach(m -> ec.cleanupCacheableData(m));
+			.forEach(ec::cleanupCacheableData);
 	}
 	
 	/**
@@ -1128,7 +1145,7 @@ public class ParForProgramBlock extends ForProgramBlock
 			.map(v -> v._name).collect(Collectors.toSet());
 		Set<String> brVars = inputs.keySet().stream()
 			.filter(v -> !retVars.contains(v))
-			.filter(v -> ec.getVariable(v).getDataType().isMatrix())
+			.filter(v -> ec.getVariable(v).getDataType().isMatrixOrFrame())
 			.filter(v -> OptimizerUtils.estimateSize(ec.getDataCharacteristics(v))< 2.14e9)
 			.collect(Collectors.toSet());
 		return brVars;
@@ -1185,7 +1202,7 @@ public class ParForProgramBlock extends ForProgramBlock
 		try
 		{
 			//create deep copies of required elements child blocks
-			ArrayList<ProgramBlock> cpChildBlocks = null;	
+			ArrayList<ProgramBlock> cpChildBlocks = null;
 			HashSet<String> fnNames = new HashSet<>();
 			if( USE_PB_CACHE )
 			{
@@ -1239,7 +1256,7 @@ public class ParForProgramBlock extends ForProgramBlock
 	 */
 	private TaskPartitioner createTaskPartitioner( IntObject from, IntObject to, IntObject incr ) 
 	{
-		TaskPartitioner tp = null;
+		TaskPartitioner tp;
 		
 		switch( _taskPartitioner ) {
 			case FIXED:
@@ -1286,7 +1303,7 @@ public class ParForProgramBlock extends ForProgramBlock
 	 */
 	private DataPartitioner createDataPartitioner(PartitionFormat dpf, PDataPartitioner dataPartitioner, ExecutionContext ec) 
 	{
-		DataPartitioner dp = null;
+		DataPartitioner dp;
 		
 		//determine max degree of parallelism
 		int numRed = OptimizerUtils.isSparkExecutionMode() ?
@@ -1307,33 +1324,41 @@ public class ParForProgramBlock extends ForProgramBlock
 		return dp;
 	}
 
-	private ResultMerge createResultMerge( PResultMerge prm, MatrixObject out, MatrixObject[] in, String fname, boolean accum, ExecutionContext ec ) 
+	private ResultMerge<?> createResultMerge( PResultMerge prm,
+		CacheableData<?> out, CacheableData<?>[] in, String fname, boolean accum, ExecutionContext ec ) 
 	{
-		ResultMerge rm = null;
+		ResultMerge<?> rm;
 		
-		//create result merge implementation (determine degree of parallelism 
-		//only for spark to avoid unnecessary spark context creation)
-		switch( prm )
-		{
-			case LOCAL_MEM:
-				rm = new ResultMergeLocalMemory( out, in, fname, accum );
-				break;
-			case LOCAL_FILE:
-				rm = new ResultMergeLocalFile( out, in, fname, accum );
-				break;
-			case LOCAL_AUTOMATIC:
-				rm = new ResultMergeLocalAutomatic( out, in, fname, accum );
-				break;
-			case REMOTE_SPARK:
-				int numMap = Math.max(_numThreads,
-					SparkExecutionContext.getDefaultParallelism(true));
-				int numRed = numMap; //equal map/reduce
-				rm = new ResultMergeRemoteSpark( out, in,
-					fname, accum, ec, numMap, numRed );
-				break;
-				
-			default:
-				throw new DMLRuntimeException("Undefined result merge: '" +prm.toString()+"'.");
+		if( out instanceof FrameObject ) {
+			rm = new ResultMergeFrameLocalMemory((FrameObject)out, (FrameObject[])in, fname, accum);
+		}
+		else if(out instanceof MatrixObject) {
+			//create result merge implementation (determine degree of parallelism 
+			//only for spark to avoid unnecessary spark context creation)
+			switch( prm )
+			{
+				case LOCAL_MEM:
+					rm = new ResultMergeLocalMemory( (MatrixObject)out, (MatrixObject[])in, fname, accum );
+					break;
+				case LOCAL_FILE:
+					rm = new ResultMergeLocalFile( (MatrixObject)out, (MatrixObject[])in, fname, accum );
+					break;
+				case LOCAL_AUTOMATIC:
+					rm = new ResultMergeLocalAutomatic( (MatrixObject)out, (MatrixObject[])in, fname, accum );
+					break;
+				case REMOTE_SPARK:
+					int numMap = Math.max(_numThreads,
+						SparkExecutionContext.getDefaultParallelism(true));
+					int numRed = numMap; //equal map/reduce
+					rm = new ResultMergeRemoteSpark( (MatrixObject)out,
+						(MatrixObject[])in, fname, accum, ec, numMap, numRed );
+					break;
+				default:
+					throw new DMLRuntimeException("Undefined result merge: '" + prm +"'.");
+			}
+		}
+		else {
+			throw new DMLRuntimeException("Unsupported result merge data: "+out.getClass().getSimpleName());
 		}
 		
 		return rm;
@@ -1437,14 +1462,15 @@ public class ParForProgramBlock extends ForProgramBlock
 			{
 				Data dat = ec.getVariable(var._name);
 				
-				if( dat instanceof MatrixObject ) //robustness scalars
+				if( dat instanceof MatrixObject | dat instanceof FrameObject )
 				{
-					MatrixObject out = (MatrixObject) dat;
-					MatrixObject[] in = Arrays.stream(results).map(vars -> 
-						vars.get(var._name)).toArray(MatrixObject[]::new);
+					CacheableData<?> out = (CacheableData<?>) dat;
+					Stream<Object> tmp = Arrays.stream(results).map(vars -> vars.get(var._name));
+					CacheableData<?>[] in = (dat instanceof MatrixObject) ?
+						tmp.toArray(MatrixObject[]::new) : tmp.toArray(FrameObject[]::new);
 					String fname = constructResultMergeFileName();
-					ResultMerge rm = createResultMerge(_resultMerge, out, in, fname, var._isAccum, ec);
-					MatrixObject outNew = USE_PARALLEL_RESULT_MERGE ?
+					ResultMerge<?> rm = createResultMerge(_resultMerge, out, in, fname, var._isAccum, ec);
+					CacheableData<?> outNew = USE_PARALLEL_RESULT_MERGE ?
 						rm.executeParallelMerge(_numThreads) :
 						rm.executeSerialMerge();
 					
@@ -1493,7 +1519,7 @@ public class ParForProgramBlock extends ForProgramBlock
 			throw new DMLRuntimeException("PARFOR: Number of executed tasks does not match the number of created tasks: tasks "+numTasks+"/"+expTasks+", iters "+numIters+"/"+expIters+".");
 	
 		if( DMLScript.STATISTICS )
-			Statistics.incrementParForMergeTime((long) time.stop());
+			ParForStatistics.incrementMergeTime((long) time.stop());
 	}
 	
 	/**
@@ -1626,9 +1652,9 @@ public class ParForProgramBlock extends ForProgramBlock
 	 */
 	private class ResultMergeWorker extends Thread
 	{
-		private LocalTaskQueue<ResultVar> _q = null;
-		private LocalVariableMap[] _refVars = null;
-		private ExecutionContext _ec = null;
+		private final LocalTaskQueue<ResultVar> _q;
+		private final LocalVariableMap[] _refVars;
+		private final ExecutionContext _ec;
 		private boolean _success = false;
 		
 		public ResultMergeWorker( LocalTaskQueue<ResultVar> q, LocalVariableMap[] results, ExecutionContext ec )
@@ -1653,18 +1679,19 @@ public class ParForProgramBlock extends ForProgramBlock
 					if( var == LocalTaskQueue.NO_MORE_TASKS ) // task queue closed (no more tasks)
 						break;
 				
-					MatrixObject out = null;
+					CacheableData<?> out = null;
 					synchronized( _ec.getVariables() ){
-						out = _ec.getMatrixObject(var._name);
+						out = _ec.getCacheableData(var._name);
 					}
 					
-					MatrixObject[] in = new MatrixObject[ _refVars.length ];
-					for( int i=0; i< _refVars.length; i++ )
-						in[i] = (MatrixObject) _refVars[i].get( var._name ); 
+					Stream<Object> tmp = Arrays.stream(_refVars).map(vars -> vars.get(var._name));
+					CacheableData<?>[] in = (out instanceof MatrixObject) ?
+						tmp.toArray(MatrixObject[]::new) : tmp.toArray(FrameObject[]::new);
+					
 					String fname = constructResultMergeFileName();
 				
-					ResultMerge rm = createResultMerge(_resultMerge, out, in, fname, var._isAccum, _ec);
-					MatrixObject outNew = null;
+					ResultMerge<?> rm = createResultMerge(_resultMerge, out, in, fname, var._isAccum, _ec);
+					CacheableData<?> outNew;
 					if( USE_PARALLEL_RESULT_MERGE )
 						outNew = rm.executeParallelMerge( _numThreads );
 					else
