@@ -19,30 +19,20 @@
 
 package org.apache.sysds.runtime.instructions.cp;
 
-import java.util.ArrayList;
-import java.util.Arrays;
-import java.util.HashMap;
-import java.util.LinkedHashMap;
-import java.util.List;
-import java.util.stream.Collectors;
-import java.util.stream.IntStream;
-
 import org.apache.commons.lang3.tuple.Pair;
 import org.apache.commons.logging.Log;
 import org.apache.commons.logging.LogFactory;
 import org.apache.sysds.api.DMLScript;
 import org.apache.sysds.common.Types.DataType;
 import org.apache.sysds.common.Types.ValueType;
+import org.apache.sysds.hops.OptimizerUtils;
 import org.apache.sysds.lops.Lop;
 import org.apache.sysds.parser.ParameterizedBuiltinFunctionExpression;
 import org.apache.sysds.parser.Statement;
 import org.apache.sysds.runtime.DMLRuntimeException;
-import org.apache.sysds.runtime.controlprogram.caching.CacheBlock;
-import org.apache.sysds.runtime.controlprogram.caching.CacheableData;
-import org.apache.sysds.runtime.controlprogram.caching.FrameObject;
-import org.apache.sysds.runtime.controlprogram.caching.MatrixObject;
-import org.apache.sysds.runtime.controlprogram.caching.TensorObject;
+import org.apache.sysds.runtime.controlprogram.caching.*;
 import org.apache.sysds.runtime.controlprogram.context.ExecutionContext;
+import org.apache.sysds.runtime.controlprogram.context.ExecutionContextFactory;
 import org.apache.sysds.runtime.data.TensorBlock;
 import org.apache.sysds.runtime.functionobjects.ParameterizedBuiltin;
 import org.apache.sysds.runtime.functionobjects.ValueFunction;
@@ -61,8 +51,13 @@ import org.apache.sysds.runtime.transform.encode.MultiColumnEncoder;
 import org.apache.sysds.runtime.transform.meta.TfMetaUtils;
 import org.apache.sysds.runtime.transform.tokenize.Tokenizer;
 import org.apache.sysds.runtime.transform.tokenize.TokenizerFactory;
+import org.apache.sysds.runtime.util.AutoDiff;
 import org.apache.sysds.runtime.util.DataConverter;
 
+import java.util.*;
+import java.util.stream.Collectors;
+import java.util.stream.IntStream;
+
 public class ParameterizedBuiltinCPInstruction extends ComputationCPInstruction {
 	private static final Log LOG = LogFactory.getLog(ParameterizedBuiltinCPInstruction.class.getName());
 	private static final int TOSTRING_MAXROWS = 100;
@@ -91,7 +86,6 @@ public class ParameterizedBuiltinCPInstruction extends ComputationCPInstruction
 	public static LinkedHashMap<String, String> constructParameterMap(String[] params) {
 		// process all elements in "params" except first(opcode) and last(output)
 		LinkedHashMap<String, String> paramMap = new LinkedHashMap<>();
-
 		// all parameters are of form <name=value>
 		String[] parts;
 		for(int i = 1; i <= params.length - 2; i++) {
@@ -150,7 +144,7 @@ public class ParameterizedBuiltinCPInstruction extends ComputationCPInstruction
 		}
 		else if(opcode.equals("transformapply") || opcode.equals("transformdecode") ||
 			opcode.equals("transformcolmap") || opcode.equals("transformmeta") || opcode.equals("tokenize") ||
-			opcode.equals("toString") || opcode.equals("nvlist")) {
+			opcode.equals("toString") || opcode.equals("nvlist") || opcode.equals("autoDiff")) {
 			return new ParameterizedBuiltinCPInstruction(null, paramsMap, out, opcode, str);
 		}
 		else if("paramserv".equals(opcode)) {
@@ -178,6 +172,13 @@ public class ParameterizedBuiltinCPInstruction extends ComputationCPInstruction
 			sores = new DoubleObject(result);
 			ec.setScalarOutput(output.getName(), sores);
 		}
+		else if(opcode.equalsIgnoreCase("autoDiff"))
+		{
+			ArrayList<Data> lineage = (ArrayList<Data>) ec.getListObject(params.get("lineage")).getData();
+			MatrixObject mo = ec.getMatrixObject(params.get("output"));
+			ListObject diffs = AutoDiff.getBackward(mo, lineage, ExecutionContextFactory.createContext());
+			ec.setVariable(output.getName(), diffs);
+		}
 		else if(opcode.equalsIgnoreCase("groupedagg")) {
 			// acquire locks
 			MatrixBlock target = ec.getMatrixInput(params.get(Statement.GAGG_TARGET));
@@ -208,29 +209,53 @@ public class ParameterizedBuiltinCPInstruction extends ComputationCPInstruction
 			String margin = params.get("margin");
 			if(!(margin.equals("rows") || margin.equals("cols")))
 				throw new DMLRuntimeException("Unspupported margin identifier '" + margin + "'.");
+			if(ec.isFrameObject(params.get("target"))) {
+				FrameBlock target = ec.getFrameInput(params.get("target"));
+				MatrixBlock select = params.containsKey("select") ? ec.getMatrixInput(params.get("select")) : null;
 
-			// acquire locks
-			MatrixBlock target = ec.getMatrixInput(params.get("target"));
-			MatrixBlock select = params.containsKey("select") ? ec.getMatrixInput(params.get("select")) : null;
+				boolean emptyReturn = Boolean.parseBoolean(params.get("empty.return").toLowerCase());
+				FrameBlock soresBlock = target.removeEmptyOperations(margin.equals("rows"), emptyReturn, select);
+				ec.setFrameOutput(output.getName(), soresBlock);
+				ec.releaseFrameInput(params.get("target"));
+				if(params.containsKey("select"))
+					ec.releaseMatrixInput(params.get("select"));
+			} else {
+				// acquire locks
+				MatrixBlock target = ec.getMatrixInput(params.get("target"));
+				MatrixBlock select = params.containsKey("select") ? ec.getMatrixInput(params.get("select")) : null;
 
-			// compute the result
-			boolean emptyReturn = Boolean.parseBoolean(params.get("empty.return").toLowerCase());
-			MatrixBlock soresBlock = target
-				.removeEmptyOperations(new MatrixBlock(), margin.equals("rows"), emptyReturn, select);
+				// compute the result
+				boolean emptyReturn = Boolean.parseBoolean(params.get("empty.return").toLowerCase());
+				MatrixBlock soresBlock = target.removeEmptyOperations(new MatrixBlock(), margin.equals("rows"), emptyReturn, select);
 
-			// release locks
-			ec.setMatrixOutput(output.getName(), soresBlock);
-			ec.releaseMatrixInput(params.get("target"));
-			if(params.containsKey("select"))
-				ec.releaseMatrixInput(params.get("select"));
+				// release locks
+				ec.setMatrixOutput(output.getName(), soresBlock);
+				ec.releaseMatrixInput(params.get("target"));
+				if(params.containsKey("select"))
+					ec.releaseMatrixInput(params.get("select"));
+			}
 		}
 		else if(opcode.equalsIgnoreCase("replace")) {
-			MatrixBlock target = ec.getMatrixInput(params.get("target"));
-			double pattern = Double.parseDouble(params.get("pattern"));
-			double replacement = Double.parseDouble(params.get("replacement"));
-			MatrixBlock ret = target.replaceOperations(new MatrixBlock(), pattern, replacement);
-			ec.setMatrixOutput(output.getName(), ret);
-			ec.releaseMatrixInput(params.get("target"));
+			if(ec.isFrameObject(params.get("target"))){
+				FrameBlock target = ec.getFrameInput(params.get("target"));
+				String pattern = params.get("pattern");
+				String replacement = params.get("replacement");
+				FrameBlock ret = target.replaceOperations(pattern, replacement);
+				ec.setFrameOutput(output.getName(), ret);
+				ec.releaseFrameInput(params.get("target"));
+			} else{
+				MatrixObject targetObj = ec.getMatrixObject(params.get("target"));
+				MatrixBlock target = targetObj.acquireRead();
+				double pattern = Double.parseDouble(params.get("pattern"));
+				double replacement = Double.parseDouble(params.get("replacement"));
+				MatrixBlock ret = target.replaceOperations(new MatrixBlock(), pattern, replacement);
+				if( ret == target ) //shallow copy (avoid bufferpool pollution)
+					ec.setVariable(output.getName(), targetObj);
+				else
+					ec.setMatrixOutput(output.getName(), ret);
+				targetObj.release();
+			}
+			
 		}
 		else if(opcode.equals("lowertri") || opcode.equals("uppertri")) {
 			MatrixBlock target = ec.getMatrixInput(params.get("target"));
@@ -279,7 +304,7 @@ public class ParameterizedBuiltinCPInstruction extends ComputationCPInstruction
 			// compute transformapply
 			MultiColumnEncoder encoder = EncoderFactory
 				.createEncoder(params.get("spec"), colNames, data.getNumColumns(), meta);
-			MatrixBlock mbout = encoder.apply(data);
+			MatrixBlock mbout = encoder.apply(data, OptimizerUtils.getTransformNumThreads());
 
 			// release locks
 			ec.setMatrixOutput(output.getName(), mbout);
@@ -376,7 +401,8 @@ public class ParameterizedBuiltinCPInstruction extends ComputationCPInstruction
 					decimal);
 			}
 			else {
-				throw new DMLRuntimeException("toString only converts matrix, tensors, lists or frames to string");
+				throw new DMLRuntimeException("toString only converts "
+					+ "matrix, tensors, lists or frames to string: "+cacheData.getClass().getSimpleName());
 			}
 			if(!(cacheData instanceof ListObject)) {
 				ec.releaseCacheableData(getParam("target"));
@@ -386,7 +412,8 @@ public class ParameterizedBuiltinCPInstruction extends ComputationCPInstruction
 		else if(opcode.equals("nvlist")) {
 			// obtain all input data objects and names in insertion order
 			List<Data> data = params.values().stream()
-				.map(d -> ec.containsVariable(d) ? ec.getVariable(d) : new StringObject(d))
+				.map(d -> ec.containsVariable(d) ? ec.getVariable(d) :
+					ScalarObjectFactory.createScalarObject(d))
 				.collect(Collectors.toList());
 			List<String> names = new ArrayList<>(params.keySet());
 
@@ -491,7 +518,7 @@ public class ParameterizedBuiltinCPInstruction extends ComputationCPInstruction
 			return Pair.of(output.getName(),
 				new LineageItem(getOpcode(), LineageItemUtils.getLineage(ec, target, meta, spec)));
 		}
-		else if (opcode.equalsIgnoreCase("nvlist")) {
+		else if (opcode.equalsIgnoreCase("nvlist") || opcode.equalsIgnoreCase("autoDiff")) {
 			List<String> names = new ArrayList<>(params.keySet());
 			CPOperand[] listOperands = names.stream().map(n -> ec.containsVariable(params.get(n)) 
 					? new CPOperand(n, ec.getVariable(params.get(n))) 
